(ns clj-in-act.ch3
  "Building blocks of Clojure"
  (:import java.util.UUID)
  (:require [clojure.repl :refer [doc]]
            [clj-in-act.ch3.listings :refer :all]))

;;; 3.1 Metadata
(def untrusted (with-meta {:command "delete-table" :subject "users"}
                          {:safe false :io true}))

(def untrusted ^{:safe false :io true} {:command "delete-table"
                                        :subject "users"})

untrusted
;=> {:command "delete-table", :subject "users"}

(def trusted {:command "delete-table" :subject "users"}) ; No metadata
;=> #'clj-in-act.ch3/trusted
(= trusted untrusted) ; But still equal
;=> true

(meta untrusted)
;=> {:safe false, :io true}
(meta trusted)
;=> nil

(def still-untrusted (assoc untrusted :complete? false))
;=> #'clj-in-act.ch3/still-untrusted
still-untrusted
;=> {:complete? false, :command "delete-table", :subject "users"}
(meta still-untrusted)
;=> {:safe false, :io true}

(defn ^{:safe true :console true
        :doc "testing metadata for functions"}
  testing-meta
  []
  (println "Hello from meta!"))

(meta testing-meta)
;=> nil

(meta (var testing-meta))
;=> {:ns #<Namespace user>,
;    :name testing-meta,
;    :file "NO_SOURCE_FILE",
;    :line 1, :arglists ([]),
;    :console true,
;    :safe true,
;    :doc "testing metadata for functions"}

;;; Java type hints

(set! *warn-on-reflection* true) ; Warn us when reflection is needed.
;=> true
(defn string-length [x] (.length x))
;Reflection warning, reference to field length can't be resolved.
;=> #'clj-in-act.ch3/string-length
(time (reduce + (map string-length (repeat 10000 "12345"))))
;"Elapsed time: 45.751 msecs"
;=> 50000
(defn fast-string-length [^String x] (.length x)) ; No reflection warning.
;=> #'clj-in-act.ch3/fast-string-length
(time (reduce + (map fast-string-length (repeat 10000 "12345"))))
;"Elapsed time: 5.788 msecs"
;=> 50000
(meta  #'fast-string-length)
;=> {:ns #<Namespace user>, :name fast-string-length, :file "NO_SOURCE_FILE",
;    :column 1, :line 1, :arglists ([x])}

;; This is how Clojure stores type hints on function arguments. Here we inspect
;; the metadata on fast-string-length, get its :arglists (list of the signatures
;; of all a function’s arities), and get the metadata on the x symbol in the argument list.
(meta (first (first (:arglists (meta #'fast-string-length)))))
;=> {:tag String}


;; `class` returns an instance of java.lang.Class representing the class of its
;; argument. Class has a method getName which returns the class name as a string.
(defn array-type [klass]
  (.getName (class (make-array klass 0))))
;=> #'clj-in-act.ch3/array-type
(array-type BigDecimal)
;=> "[Ljava.math.BigDecimal;"
(def bigdec-arr
  ^"[Ljava.math.BigDecimal;" ; Yep, type hint with a string name!
  ;; `into-array` returns a Java array of a specified type filled with items from
  ;; a Clojure collection.
  (into-array BigDecimal [1.0M]))


;;; 3.2 Java exceptions: try and throw

(defn average [numbers]
  (let [total (apply + numbers)]
    (/ total (count numbers))))

#_(average [])
;ArithmeticException Divide by zero  clojure.lang.Numbers.divide (Numbers.java:156)

(defn safe-average [numbers]
  (let [total (apply + numbers)]
    (try
      (/ total (count numbers))
      (catch ArithmeticException e ; The exception object is bound to 'e'.
        (println "Divided by zero!")
        0))))
;=> #'clj-in-act.ch3/safe-average
(safe-average [])
;Divided by zero!
;=> 0


(try
  (print "Attempting division... ")
  (/ 1 0)
  ;; RuntimeException is a superclass of ArithmeticException.
  (catch RuntimeException e "Runtime exception!")
  (catch ArithmeticException e "DIVIDE BY ZERO!")
  ;; Throwable is the most generic possible exception type in Java. All
  ;; exception types are subclasses of java.lang.Throwable.
  (catch Throwable e "Unknown exception encountered!")
  ;; The finally clause is always executed, even if no exception is thrown.
  (finally
    (println "done.")))
;Attempting division... done.
;; Not "DIVIDE BY ZERO!" as you expected; see discussion in section 3.2.
;=> "Runtime exception!"

#_(try
  (print "Attempting division... ")
  (/ 1 0)
  (finally
    (println "done.")))
;; If no catch clause matches the exception is thrown as normal, but finally is still executed.
;Attempting division... done.
;ArithmeticException Divide by zero  clojure.lang.Numbers.divide (Numbers.java:156)


;;; 3.3 Functions

(defn total-cost [item-cost number-of-items]
  (* item-cost number-of-items))

(def total-cost (fn [item-cost number-of-items]
                  (* item-cost number-of-items)))

(defn total-cost
  "return line-item total of the item and quantity provided"
  [item-cost number-of-items]
  (* item-cost number-of-items))

(meta #'total-cost)
;=> {:ns #<Namespace clj-in-act.ch3>, :name total-cost, :file "NO_SOURCE_FILE", :column 1,
;    :line 1, :arglists ([item-cost number-of-items]),
;    :doc "return line-item total of the item and quantity provided"}

(doc total-cost)
;-------------------------
;clj-in-act.ch3/total-cost
;([item-cost number-of-items])
;return line-item total of the item and quantity provided
;=> nil


(meta (defn myfn-attr-map {:a 1} [])) ; Metadata without using ^{}.
;=> {:ns #<Namespace clj-in-act.ch3>, :name myfn-attr-map, :file "NO_SOURCE_FILE",
;    :column 7, :line 1, :arglists ([]), :a 1}
(meta (defn ^{:a 1} myfn-metadata [])) ; Notice the ^{} before the name.
;=> {:ns #<Namespace clj-in-act.ch3>, :name myfn-metadata, :file "NO_SOURCE_FILE",
;    :column 7, :line 1, :arglists ([]), :a 1}
(meta (defn ^{:a 1} myfn-both {:a 2 :b 3} [])) ; Metadata merged together!
;=> {:ns #<Namespace clj-in-act.ch3>, :name myfn-both, :file "NO_SOURCE_FILE",
;    :column 7, :line 1, :arglists ([]), :a 2, :b 3}

;; Metadata key-value pairs from the different metadata sources overwrite each
;; other from left to right.
(meta (defn ^{:a 1 :doc "doc 1"} myfn-redundant-docs "doc 2" {:a 2 :b 3 :doc "doc 3"} []))
;=> {:ns #<Namespace clj-in-act.ch3>, :name myfn-redundant-docs,
;    :file "NO_SOURCE_FILE", :column 7, :line 1, :arglists ([]),
;    :a 2, :b 3, :doc "doc 3"}


(defn item-total [price quantity discount-percentage]
  {:pre [(> price 0) (> quantity 0)]
   :post [(> % 0)]}
  (->> (/ discount-percentage 100)
       (- 1)
       (* price quantity)
       float))


(item-total 100 2 0)
;=> 200.0
(item-total 100 2 10)
;=> 180.0


#_(item-total 100 -2 10)
;AssertionError Assert failed: (> quantity 0)  clj-in-act.ch3/item-total


(defn total-cost
  ([item-cost number-of-items]
   (* item-cost number-of-items))
  ([item-cost]
   (total-cost item-cost 1)))

(defn total-all-numbers [& numbers]
  (apply + numbers))

(defn many-arities
  ([]             0)
  ([a]            1)
  ([a b c]        3)
  ([a b c & more] "variadic"))
;=> #'clj-in-act.ch3/many-arities
(many-arities)
;=> 0
(many-arities "one argument")
;=> 1
#_(many-arities "two" "arguments")
;ArityException Wrong number of args (2) passed to: clj-in-act.ch3/many-arities
; clojure.lang.AFn.throwArity (AFn.java:429)
(many-arities "three" "argu-" "ments")
;=> 3
(many-arities "many" "more" "argu-" "ments")
;=> "variadic"


(defn count-down [n]
  (when-not (zero? n)
    (when (zero? (rem n 100))      ; `rem` means remainder of division
      (println "count-down:" n))
    (count-down (dec n))))        ; `dec` means decrement by one


(defn count-downr [n]
  (when-not (zero? n)
    (if (zero? (rem n 100))
      (println "count-down:" n))
    (recur (dec n))))


;;; Higher-order functions

(def bools [true true true false false])
;=> #'clj-in-act.ch3/bools
(every? true? bools)  ; true? is a predicate function.
;=> false


(some (fn [p] (= "rob" p)) ["kyle" "siva" "rob" "celeste"])
;=> true returns true


(def two (constantly 2)) ; same as (def two (fn [& more] 2))
; or      (defn two [& more] 2)
;=> #'clj-in-act.ch3/two
(two 1)
;=> 2
(two :a :b :c)
;=> 2


(defn greater? [x y]
  (> x y))

(greater? 10 5)
;=> true
(greater? 10 20)
;=> false

(def smaller? (complement greater?))

(smaller? 10 5)
;=> false

(smaller? 10 20)
;=> true

(def opp-zero-str (comp str not zero?))

(opp-zero-str 0)
;=> "false"

(opp-zero-str 1)
;=> "true"

(defn above-threshold? [threshold number]
  (> number threshold))

(filter (fn [x] (above-threshold? 5 x)) [ 1 2 3 4 5 6 7 8 9])
;=> (6 7 8 9)

(filter (partial above-threshold? 5) [ 1 2 3 4 5 6 7 8 9])
;=> (6 7 8 9)

(defn slow-calc [n m]
  (Thread/sleep 1000)
  (* n m))

(time (slow-calc 5 7))
;"Elapsed time: 1000.097 msecs"
;=> 35

(def fast-calc (memoize slow-calc))

(time (fast-calc 5 7))
;"Elapsed time: 0.035 msecs"
;=> 35


(map username users)
;=> ("kyle" "zak" "rob")
#_(sort *1) ; Remember, *1 means “last result returned by the REPL”
;=> ("kyle" "rob" "zak")


(sort-by username users)
;=> ({:member-since "2009-04-16", :username "kyle", …}
;    {:member-since "2009-03-30", :username "rob",  …}
;    {:member-since "2009-02-01", :username "zak",  …})

(poorest-first users)
;=> ({:username "zak",  :balance  12.95M, …}
;    {:username "rob",  :balance  98.50M, …}
;    {:username "kyle", :balance 175.00M, …})

(map lastname-firstname users)
;=> (["Smith" "Kyle"] ["Jones" "Zackary"] ["Jones" "Robert"])
(sort *1)
;=> (["Jones" "Robert"] ["Jones" "Zackary"] ["Smith" "Kyle"])

(last-then-firstname users)
;=> ({:lastname "Jones", :firstname "Robert",  :username "rob",  …}
;    {:lastname "Jones", :firstname "Zackary", :username "zak",  …}
;    {:lastname "Smith", :firstname "Kyle",    :username "kyle", …})

(def total-cost
  (fn [item-cost number-of-items]
    (* item-cost number-of-items)))

(map (fn [user] (user :member-since)) users)
;=> ("2009-04-16" "2009-02-01" "2009-03-30")

(map #(% :member-since) users) ; Same result!
;=> ("2009-04-16" "2009-02-01" "2009-03-30")


(#(vector %&) 1 2 3 4 5)
;=> [(1 2 3 4 5)]
(#(vector % %&) 1 2 3 4 5)
;=> [1 (2 3 4 5)]
(#(vector %1 %2 %&) 1 2 3 4 5)
;=> [1 2 (3 4 5)]
(#(vector %1 %2 %&) 1 2)
;=> [1 2 nil]


(def person {:username "zak"
             :balance 12.95
             :member-since "2009-02-01"})

(person :username)
;=> "zak"

(:username person)
;=> "zak"

(map #(% :member-since) users)
;=> ("2009-04-16" "2009-02-01" "2009-03-30")

(map :member-since users)
;=> ("2009-04-16" "2009-02-01" "2009-03-30")

(:login person)
;=> nil
(:login person :not-found)
;=> :not-found


(def expense {'name "Snow Leopard" 'cost 29.95M})
;=> #'clj-in-act.ch3/expense
(expense 'name)
;=> "Snow Leopard"
('name expense)
;=> "Snow Leopard"
('vendor expense)
;=> nil
('vendor expense :absent)
;=> :absent

(person :login :not-found)
;=> :not-found

(def names ["kyle" "zak" "rob"])
;=> #'clj-in-act.ch3/names
(names 1)
;=> "zak"


#_(names 10)
;IndexOutOfBoundsException   clojure.lang.PersistentVector.arrayFor (PersistentVector.java:107)
#_(names 10 :not-found)
;ArityException Wrong number of args (2) passed to: PersistentVector  clojure.lang.AFn.throwArity (AFn.java:429)


;;; 3.4 Scope

(def MAX-CONNECTIONS 10)

(def RABBITMQ-CONNECTION)

(def ^:dynamic RABBITMQ-CONNECTION)

(defn new-connection []
  ;; Return a RABBITMQ connection
  )

(binding [RABBITMQ-CONNECTION (new-connection)]
  (
    ;; do something here with RABBITMQ-CONNECTION
    ))

(def ^:dynamic *db-host* "localhost")

(defn expense-report [start-date end-date]
  (println *db-host*)) ;; can do real work

(binding [*db-host* "production"]
  (expense-report "2010-01-01" "2010-01-07"))

(def ^:dynamic *factor* 10)
(defn multiply [x]
  (* x *factor*))

(map multiply [1 2 3 4 5])

(binding [*factor* 20]
  (map multiply [1 2 3 4 5]))

(binding [*factor* 20]
  (doall (map multiply [1 2 3 4 5])))

(let [x 10
      y 20]
  (println "x, y:" x "," y))

(defn upcased-names [names]
  (let [up-case (fn [name] (.toUpperCase name))]
    (map up-case names)))
;=> #'clj-in-act.ch3/upcased-names
(upcased-names ["foo" "bar" "baz"])
;=> ("FOO" "BAR" "BAZ")

(def ^:dynamic *factor* 10)
(binding [*factor* 20]
  (println *factor*)
  (doall (map multiply [1 2 3 4 5])))

(let [*factor* 20]
  (println *factor*)
  (doall (map multiply [1 2 3 4 5])))


;;; 3.6 Destructuring

(defn describe-salary [person]
  (let [first  (:first-name person)
        last   (:last-name  person)
        annual (:salary     person)]
    (println first last "earns" annual)))

(defn describe-salary-2 [{first  :first-name
                          last   :last-name
                          annual :salary}]
  (println first last "earns" annual))

(defn print-amounts [[amount-1 amount-2]]
  (println "amounts are:" amount-1 "and" amount-2))
(print-amounts [10.95 31.45])
;amounts are: 10.95 and 31.45

(defn print-amounts-multiple [[amount-1 amount-2 & remaining]]
  (println "Amounts are:" amount-1 "," amount-2 "and" remaining))

(print-amounts-multiple [10.95 31.45 22.36 2.95])
;Amounts are: 10.95 , 31.45 and (22.36 2.95)

(defn print-all-amounts [[amount-1 amount-2 & remaining :as all]]
  (println "Amounts are:" amount-1 "," amount-2 "and" remaining)
  (println "Also, all the amounts are:" all))

(print-all-amounts [10.95 31.45 22.36 2.95])
;Amounts are: 10.95 , 31.45 and (22.36 2.95)
;Also, all the amounts are: [10.95 31.45 22.36 2.95]

(defn print-first-category [[[category amount] & _ ]]
  (println "First category  was:" category)
  (println "First amount was:" amount))

(def expenses [[:books 49.95] [:coffee 4.95] [:caltrain 2.25]])
(print-first-category expenses)
;First category  was: :books
;First amount was: 49.95

(defn describe-salary-3 [{first  :first-name
                          last   :last-name
                          annual :salary
                          bonus  :bonus-percentage
                          :or {bonus 5}}]
  (println first last "earns" annual "with a" bonus "percent bonus"))


(def a-user {:first-name       "pascal"
             :last-name        "dylan"
             :salary           85000
             :bonus-percentage 20})
(describe-salary-3 a-user)
;pascal dylan earns 85000 with a 20 percent bonus

(def another-user {:first-name "basic"
                   :last-name  "groovy"
                   :salary     70000})
(describe-salary-3 another-user)
;basic groovy earns 70000 with a 5 percent bonus

(defn describe-person [{first  :first-name
                        last   :last-name
                        bonus  :bonus-percentage
                        :or {bonus 5}
                        :as p}]
  (println "Info about" first last "is:" p)
  (println "Bonus is:" bonus "percent"))

(def third-user {:first-name "lambda"
                 :last-name   "curry"
                 :salary      95000})
(describe-person third-user)

;Info about lambda curry is: {:first-name lambda,
;                             :last-name curry,
;                             :salary 95000}
;Bonus is: 5 percent


(defn greet-user [{:keys [first-name last-name]}]
  (println "Welcome," first-name last-name))


(def roger {:first-name "roger" :last-name "mann" :salary 65000})
(greet-user roger)
;Welcome, roger mann

;;; 3.7 Reader Literals


;;; guid reader is defined in chj-in-act.ch3.reader
;;; See file data_readers.clj in the src directory.



