(ns clj-in-act.ch4
  "Multimethod polymorphism")

;;; 4.1 Polymorphism and its kinds

;;; Ad hoc Polymorphism

(defn ad-hoc-type-namer [thing]
  ;; Call `type` on the thing argument to return the type of thing.
  (condp = (type thing)
    ;; Then we dispatch for an implementation of the function appropriate to the
    ;; type. In this case, the type-specific implementation is simply the string
    ;; "string".
    java.lang.String              "string"
    clojure.lang.PersistentVector "vector"))
;=> #'clj-in-act.ch4/ad-hoc-type-namer
(ad-hoc-type-namer "I'm a string")
;=> "string"
(ad-hoc-type-namer [])
;=> "vector"

;; If there’s a type this function doesn’t know how to handle, it throws an exception.
#_(ad-hoc-type-namer {})
;IllegalArgumentException No matching clause: class clojure.lang.PersistentArrayMap
; clj-in-act.ch4/ad-hoc-type-namer (NO_SOURCE_FILE:2)


;; Pull the implementations out into a separate redef-able map.
(def type-namer-implementations
  {java.lang.String              (fn [thing] "string")
   clojure.lang.PersistentVector (fn [thing] "vector")})
;=> #'clj-in-act.ch4/type-namer-implementations
(defn open-ad-hoc-type-namer [thing]
  ;; Use a dispatch value as the key to the implementation map.
  (let [dispatch-value (type thing)]
    (if-let [implementation
             (get type-namer-implementations dispatch-value)]
      ;; If we find an implementation for the dispatch value, use it and return
      ;; the result.
      (implementation thing)
      ;; Otherwise throw an exception just like before.
      (throw (IllegalArgumentException.
               (str "No implementation found for " dispatch-value))))))
;=> #'clj-in-act.ch4/open-ad-hoc-type-namer
(open-ad-hoc-type-namer "I'm a string")
;=> "string"
(open-ad-hoc-type-namer []) ; But it still doesn’t understand maps.
;=> "vector"
#_(open-ad-hoc-type-namer {}) ; But it still doesn’t understand maps.
;IllegalArgumentException No implementation found for class clojure.lang.PersistentArrayMap  clj-in-act.ch4/open-ad-hoc-type-namer (NO_SOURCE_FILE:5)
;; So we redefine the implementation map, adding a case for PersistentArrayMap.
(def type-namer-implementations
  (assoc type-namer-implementations
    clojure.lang.PersistentArrayMap (fn [thing] "map")))
;=> #'clj-in-act.ch4/type-namer-implementations
;; And voila, open-ad-hoc-type-namer now understands maps!
(open-ad-hoc-type-namer {})
;=> "map"


;;; Subtype Polymorphism

;; First attempt with ad hoc polymorphism checks each type explicitly.
(defn map-type-namer [thing]
  (condp = (type thing)
    clojure.lang.PersistentArrayMap "map"
    clojure.lang.PersistentHashMap  "map")) ; Notice the duplicate implementation.
;=> #'clj-in-act.ch4/map-type-namer
(map-type-namer (hash-map))
;=> "map"
(map-type-namer (array-map))
;=> "map"
#_(map-type-namer (sorted-map)) ; There was a case we missed!
;IllegalArgumentException No matching clause: class clojure.lang.PersistentTreeMap  com.gentest.ConcreteClojureClass/map-type-namer (NO_SOURCE_FILE:2)

(defn subtyping-map-type-namer [thing] ; Second attempt using subtype polymorphism.
  (cond
    ;; We use instance? to query Java’s class hierarchy. APersistentMap is the
    ;; Java superclass of all map-like things in Clojure.
    (instance? clojure.lang.APersistentMap thing)   "map"
    :else (throw (IllegalArgumentException.
                   (str "No implementation found for ") (type thing)))))
;=> #'clj-in-act.ch4/subtyping-map-type-namer
(subtyping-map-type-namer (hash-map))
;=> "map"
(subtyping-map-type-namer (array-map))
;=> "map"
;; This function now works for anything map-like, using a single implementation.
;; It even works for types that don’t exist yet! (As long as they will extend
;; APersistentMap.)
(subtyping-map-type-namer (sorted-map)) ;#F
;=> "map"


;;; 4.2 Polymorphism using multimethods


;;; Without multimethods

(def example-user {:login "rob" :referrer "mint.com" :salary 100000})
;=> #'clj-in-act.ch4/example-user
(defn fee-amount [percentage user]
  ;; We are using BigDecimal because we’re dealing with money.
  (with-precision 16 :rounding HALF_EVEN
                     (* 0.01M percentage (:salary user))))
;=> #'clj-in-act.ch4/fee-amount
(defn affiliate-fee [user]
  (case (:referrer user)
    "google.com" (fee-amount 0.01M user)
    "mint.com"   (fee-amount 0.03M user)
    (fee-amount 0.02M user)))
;=> #'clj-in-act.ch4/affiliate-fee
(affiliate-fee example-user)
;=> 30.0000M

;;; Ad hoc polymorphism using multimethods

;; Like `(defn affiliate-fee-dispatch [user] (:referrer user))`
(defmulti affiliate-fee (fn [user] (:referrer user)))
;=> #'clj-in-act.ch4/affiliate-fee
;; Like a clause of the case: (case (affiliate-fee-dispatch user)
;;                              "mint.com" (fee-amount 0.03M user))
(defmethod affiliate-fee "mint.com" [user]
  (fee-amount 0.03M user))
;; Notice defmethod does not return a var.
;=> #<MultiFn clojure.lang.MultiFn@8b3cf12>
(defmethod affiliate-fee "google.com" [user]
  (fee-amount 0.01M user))
;=> #<MultiFn clojure.lang.MultiFn@8b3cf12>
;; Our default case. Multimethods use :default as a dispatch value when no other
;; defmethod matches.
(defmethod affiliate-fee :default [user]
  (fee-amount 0.02M user))
;=> #<MultiFn clojure.lang.MultiFn@8b3cf12>
(affiliate-fee example-user)
;=> 30.0000M

;; Important note for REPL use: a defmulti cannot be redefined (it uses defonce),
;; and if it is redefined all its defmethod need to be redeclared, too.

;; `defmulti` returning `nil` means it was not redefined: you need to unmap its
;; var first.
(defmulti affiliate-fee :referrer :default "*")
;=> nil

(ns-unmap 'clj-in-act.ch4 'affiliate-fee) ; This is how you do this at the REPL.
;=> nil
;; We add the option :default "*". Also notice we can use a plain keyword
;; :referrer as the dispatch function: this is a common idiom where the argument
;; is always a map.
(defmulti affiliate-fee :referrer :default "*")
;=> #'clj-in-act.ch4/affiliate-fee
;; Now instead of :default, the default value is "*".
(defmethod affiliate-fee "*" [user]
  (fee-amount 0.02M user))
;=> #<MultiFn clojure.lang.MultiFn@7eafa7a7>
;; Notice because of the redef our "mint.com" case is gone (we get 20 not 30),
;; but the default case still works.
(affiliate-fee example-user)
;=> 20.0000M


;; First let’s add back the methods we lost after the redef.
(defmethod affiliate-fee "mint.com" [user]
  (fee-amount 0.03M user))
;=> #<MultiFn clojure.lang.MultiFn@7eafa7a7>
(defmethod affiliate-fee "google.com" [user]
  (fee-amount 0.01M user))
;=> #<MultiFn clojure.lang.MultiFn@7eafa7a7>
(methods affiliate-fee) ; Get the full dispatch map for a multimethod.
;=> {"mint.com"   #<user$eval8117$fn__8118 user$eval8117$fn__8118@36e433f1>,
;    "*"          #<user$eval6380$fn__6381 user$eval6380$fn__6381@2ebfd12f>,
;    "google.com" #<user$eval8123$fn__8124 user$eval8123$fn__8124@2ad71f9f>}

;; Get the implementation for a dispatch value.
(get-method affiliate-fee "mint.com")
;=> #<user$eval8117$fn__8118 user$eval8117$fn__8118@36e433f1>
(get (methods affiliate-fee) "example.org")
;=> nil
;; `get-method` is not just get: it understands the dispatch logic.
(get-method affiliate-fee "example.org")
;; The "*" implementation.
;=> #<user$eval6380$fn__6381 user$eval6380$fn__6381@2ebfd12f>
;; Entries in the dispatch map are normal functions.
((get-method affiliate-fee "mint.com") example-user)
;=> 30.0000M


;;; Multiple dispatch

(def user-1 {:login "rob"     :referrer "mint.com"   :salary 100000
             :rating :rating/bronze})
(def user-2 {:login "gordon"  :referrer "mint.com"   :salary 80000
             :rating :rating/silver})
(def user-3 {:login "kyle"    :referrer "google.com" :salary 90000
             :rating :rating/gold})
(def user-4 {:login "celeste" :referrer "yahoo.com"  :salary 70000
             :rating :rating/platinum})


(defn fee-category [user]
  [(:referrer user) (:rating user)])
;=> #'clj-in-act.ch4/fee-category
(map fee-category [user-1 user-2 user-3 user-4])
;=> (["mint.com"   :rating/bronze]
;    ["mint.com"   :rating/silver]
;    ["google.com" :rating/gold]
;    ["yahoo.com"  :rating/platinum])


;; Our multimethod uses our fee category function to create dispatch values.
(defmulti profit-based-affiliate-fee fee-category)

(defmethod profit-based-affiliate-fee ["mint.com" :rating/bronze]
  [user] (fee-amount 0.03M user))

(defmethod profit-based-affiliate-fee ["mint.com" :rating/silver]
  [user] (fee-amount 0.04M user))

(defmethod profit-based-affiliate-fee ["mint.com" :rating/gold]
  [user] (fee-amount 0.05M user))

(defmethod profit-based-affiliate-fee ["mint.com" :rating/platinum]
  [user] (fee-amount 0.05M user))

(defmethod profit-based-affiliate-fee ["google.com" :rating/gold]
  [user] (fee-amount 0.03M user))

(defmethod profit-based-affiliate-fee ["google.com" :rating/platinum]
  [user] (fee-amount 0.03M user))

;; Even with multiple dispatch, the default value is a always a single value,
;; not [:default :default]
(defmethod profit-based-affiliate-fee :default
  [user] (fee-amount 0.02M user))


(map profit-based-affiliate-fee [user-1 user-2 user-3 user-4])
;=> (30.0000M  ; mint.com bronze gets 0.03% of $100000
;    32.0000M  ; mint.com silver gets 0.04% of $80000
;    27.0000M  ; google.com gold gets 0.04% of $90000
;    14.0000M) ; yahoo.com platinum gets the default 0.02% of $70000



;;; Subtype polymorphism using multimethods

;; The nil return value is a hint that mutation is occurring.
(derive :rating/bronze :rating/basic)
(derive :rating/silver :rating/basic)
(derive :rating/gold :rating/premier)
(derive :rating/platinum :rating/premier)
;; We use :rating/ANY as the root type for all ratings.
(derive :rating/basic :rating/ANY)
(derive :rating/premier :rating/ANY)


(isa? :rating/gold :rating/premier)
;=> true
(isa? :rating/gold :rating/ANY) ; isa? understands transitive relationships, too.
;=> true
(isa? :rating/ANY :rating/premier)
;=> false
(isa? :rating/gold :rating/gold) ; Types are always kinds of themselves.
;=> true
;; parents returns a set (not a single item) because it’s legal to have multiple
;; parents!
(parents :rating/premier)
;=> #{:rating/ANY}
(ancestors :rating/gold) ; ancestors returns parents and their parents, etc.
;=> #{:rating/ANY :rating/premier}
(descendants :rating/ANY) ; descendants returns children and their children, etc.
;=> #{:rating/basic :rating/bronze :rating/gold :rating/premier :rating/silver :rating/platinum}


(defmulti greet-user :rating)
;=> #'clj-in-act.ch4/greet-user
(defmethod greet-user :rating/basic [user]
  (str "Hello " (:login user) \.))
;=> #<MultiFn clojure.lang.MultiFn@d81fe85>
(defmethod greet-user :rating/premier [user]
  (str "Welcome, " (:login user) ", valued affiliate member!"))
;=> #<MultiFn clojure.lang.MultiFn@d81fe85>
(map greet-user [user-1 user-2 user-3 user-4])
;=> ("Hello rob."
;    "Hello gordon."
;    "Welcome, kyle, valued affiliate member!"
;    "Welcome, celeste, valued affiliate member!")


;; First we remove our duplicate methods.
(remove-method profit-based-affiliate-fee ["mint.com" :rating/gold])
;=> #<MultiFn clojure.lang.MultiFn@5d98c6f3>
(remove-method profit-based-affiliate-fee ["mint.com" :rating/platinum])
;=> #<MultiFn clojure.lang.MultiFn@5d98c6f3>
(remove-method profit-based-affiliate-fee ["google.com" :rating/gold])
;=> #<MultiFn clojure.lang.MultiFn@5d98c6f3>
(remove-method profit-based-affiliate-fee ["google.com" :rating/platinum])
;=> #<MultiFn clojure.lang.MultiFn@5d98c6f3>
;; Now add two methods matching :rating/premier to replace our previous four.
(defmethod profit-based-affiliate-fee ["mint.com" :rating/premier]
  [user] (fee-amount 0.05M user))
;=> #<MultiFn clojure.lang.MultiFn@5d98c6f3>
(defmethod profit-based-affiliate-fee ["google.com" :rating/premier]
  [user] (fee-amount 0.03M user))
;=> #<MultiFn clojure.lang.MultiFn@5d98c6f3>
;; Check if we get the same results…
(map profit-based-affiliate-fee [user-1 user-2 user-3 user-4])
;=> (30.0000M 32.0000M 27.0000M 14.0000M) ; …and we do!


(defmulti size-up (fn [observer observed]
                    [(:rating observer) (:rating observed)]))
;=> #'clj-in-act.ch4/size-up
;; Everyone should look scrawny to platinum users.
(defmethod size-up [:rating/platinum :rating/ANY] [_ observed]
  (str (:login observed) " seems scrawny."))
;=> #<MultiFn clojure.lang.MultiFn@50e0a48d>
;; And platinum users should look imposing to everyone.
(defmethod size-up [:rating/ANY :rating/platinum] [_ observed]
  (str (:login observed) " shimmers with an unearthly light."))
;=> #<MultiFn clojure.lang.MultiFn@50e0a48d>
;; But what happens if a platinum user looks at another platinum user?
#_(size-up {:rating :rating/platinum} user-4)
;IllegalArgumentException Multiple methods in multimethod 'size-up' match dispatch value: [:rating/platinum :rating/platinum] -> [:rating/ANY :rating/platinum] and [:rating/platinum :rating/ANY], and neither is preferred  clojure.lang.MultiFn.findAndCacheBestMethod (MultiFn.java:182)


;; When size-up can’t decide between these two dispatch values, prefer the first
;; one.
(prefer-method size-up [:rating/ANY :rating/platinum]
               [:rating/platinum :rating/ANY])
;=> #<MultiFn clojure.lang.MultiFn@50e0a48d>
(size-up {:rating :rating/platinum} user-4) ; Now Celeste shimmers!
;=> "celeste shimmers with an unearthly light."
(prefers size-up) ; prefers shows all a multimethod’s preferences.
;=> {[:rating/ANY :rating/platinum] #{[:rating/platinum :rating/ANY]}}


(def myhier (make-hierarchy))
;=> #'clj-in-act.ch4/myhier
myhier ; Hierarchies are just ordinary maps!
;=> {:parents {}, :descendants {}, :ancestors {}}
(derive myhier :a :letter) ; Three-arg derive does not return nil.
;=> {:parents {:a #{:letter}}, :ancestors {:a #{:letter}},
;    :descendants {:letter #{:a}}}
myhier ; Because unlike two-arg derive it does not mutate!
;=> {:parents {}, :descendants {}, :ancestors {}}
(def myhier (-> myhier              ; So we need to redef to “mutate” the hierarchy.
                (derive :a :letter) ; Notice keywords without namespaces.
                (derive :b :letter)
                (derive :c :letter)))
;=> #'clj-in-act.ch4/myhier
(isa? myhier :a :letter) ; isa? can also take a hierarchy.
;=> true
(parents myhier :a) ; As can parents, ancestors, descendants, and underive.
;=> #{:letter}
;; defmulti has a :hierarchy option, which takes a var.
(defmulti letter? identity :hierarchy #'myhier)
;=> #'clj-in-act.ch4/letter?
(defmethod letter? :letter [_] true)
;=> #<MultiFn clojure.lang.MultiFn@17c26ef7>
#_(letter? :d)
;IllegalArgumentException No method in multimethod 'letter?' for dispatch value: :d  clojure.lang.MultiFn.getFn (MultiFn.java:160)
(def myhier (derive myhier :d :letter)) ; Taking a var allows mutation.
;=> #'clj-in-act.ch4/myhier
(letter? :d)
;=> true
