(ns clj-in-act.ch6)

;;; 6.4 Refs

(def all-users (ref {}))
;=> clj-in-act.ch6/all-users
(deref all-users)
;=> {}
@all-users
;=> {}
all-users
;=> #<Ref@227e9896: {}>

#_(ref-set all-users {})
;IllegalStateException No transaction running  clojure.lang.LockingTransaction.getEx
; (LockingTransaction.java:208)

(dosync
  (ref-set all-users {}))
;=> {}

(defn new-user [id login monthly-budget]
  {:id             id
   :login          login
   :monthly-budget monthly-budget
   :total-expenses 0})

(defn add-new-user [login budget-amount]
  (dosync
    (let [current-number (count @all-users)
          user (new-user (inc current-number) login budget-amount)]
      (alter all-users assoc login user))))

(add-new-user "amit" 1000000)
;=> {"amit" {:id 2, :login "amit", :monthly-budget 1000000, :total-expenses 0}}

(add-new-user "deepthi" 2000000)
;=> {"deepthi" {:id 3, :login "deepthi", :monthly-budget 2000000,
;               :total expenses 0},
;    "amit"    {:id 2, :login "amit", :monthly-budget 1000000,
;               :total-expenses 0}}


;;; 6.5 Agents

(def total-cpu-time (agent 0))
;=> clj-in-act.ch6/total-cpu-time
(deref total-cpu-time)
;=> 0
@total-cpu-time
;=> 0
(send total-cpu-time + 700)
;=> #<Agent@12518b85: 700>
@total-cpu-time
;=> 700

(def bad-agent (agent 10))

(send bad-agent / 0)
;=> #<Agent@125b9ec1 FAILED: 10>

(deref bad-agent)
;=> 10

#_(send bad-agent / 2)
;ArithmeticException Divide by zero  clojure.lang.Numbers.divide (Numbers.java:156)

(agent-error bad-agent)
;=> #<ArithmeticException java.lang.ArithmeticException: Divide by zero>

(let [e (agent-error bad-agent)
      st (.getStackTrace e)]
  (println (.getMessage e))
  (println (clojure.string/join "\n" st)))

(clear-agent-errors bad-agent)

;;; 6.6 Atoms

(def total-rows (atom 0))

(deref total-rows)
;=> 0
@total-rows
;=> 0

(reset! total-rows 100)
;=> 100
(swap! total-rows + 100)
;=> 200

(compare-and-set! total-rows 100 250)
;=> false
@total-rows
;=> 200

(compare-and-set! total-rows 200 250)
;=> true
@total-rows
;=> 250

;;; 6.7 Vars

(def hbase-master "localhost")

(def  ^:dynamic *hbase-master* "localhost")
;=> #'clj-in-act.ch6/*hbase-master*
(println "Hbase-master is:" *hbase-master*)
;Hbase-master is: localhost
;=> nil


(def ^:dynamic *rabbitmq-host*)
;=> #'clj-in-act.ch6/*rabbitmq-host*
(println "RabbitMQ host is:" *rabbitmq-host*)
;RabbitMQ host is: #<Unbound Unbound: #'clj-in-act.ch6/*rabbitmq-host*>
;=> nil
(bound? #'*rabbitmq-host*) ; You can test if a var is bound using the bound? function.
;=> false


(def ^:dynamic *mysql-host*)

(defn db-query [db]
  (binding [*mysql-host* db]
    (count *mysql-host*)))

(def mysql-hosts ["test-mysql" "dev-mysql" "staging-mysql"])


(pmap db-query mysql-hosts)
;=> (10 9 13)


;;; 6.8 State and its unified access model

;;; Creating

(def a-ref (ref 0))
(def an-agent (agent 0))
(def an-atom (atom 0))

;;; Reading

(deref a-ref)
@a-ref

(deref an-agent)
@an-agent

(deref an-atom)
@an-atom

;;; Mutation

(def adi (atom 0))
(defn on-change [the-key the-ref old-value new-value]
  (println "Hey, seeing change from" old-value "to" new-value))
(add-watch adi :adi-watcher on-change)


@adi
;=> 0
(swap! adi inc)
;Hey, seeing change from 0 to 1
;=> 1

(remove-watch adi :adi-watcher)


;;; 6.10 Futures and promises

(defn long-calculation [num1 num2]
  (Thread/sleep 5000)
  (* num1 num2))

(defn long-run []
  (let [x (long-calculation 11 13)
        y (long-calculation 13 17)
        z (long-calculation 17 19)]
    (* x y z)))

#_(time (long-run))
;"Elapsed time: 14998.165 msecs"
;=> 10207769

(defn fast-run []
  ;; future creates a thread that will run long-calculation without blocking
  ;; the current thread.
  (let [x (future (long-calculation 11 13))
        ;; The futures x, y, and z can potentially all run in parallel (if you
        ;; have enough cores), meaning they could all be finished in five
        ;; seconds total.
        y (future (long-calculation 13 17))
        z (future (long-calculation 17 19))]
    ;; Each dereference here will block in the order listed. The amount of time
    ;; the whole expression blocks will be at least five seconds, but could be
    ;; as much as 15 if none of the futures can be run in parallel (e.g. on a
    ;; single-core machine).
    (* @x @y @z)))


#_(time (fast-run))
;"Elapsed time: 5000.078 msecs"
;=> 10207769


(def p (promise))

;;; WARNING: Don’t dereference the promise at the REPL; it will block and you’ll
;;; have no way to unblock it!
#_(def value (deref p))

;; Promises always start out with no value, so promise is a no-argument function.
(let [p (promise)]
  (future (Thread/sleep 5000)
    ;; We run this future only for side-effects (the deliver), so we don’t bind
    ;; it to a variable.
    (deliver p :done))
  ;; Dereferencing the promise will block 5 seconds, then return the :done value
  ;; after the promise receives it.
  @p)
;=> :done
