(ns clj-in-act.ch7
  "Evolving Clojure through macros")

;;; 7.1 Macro basics

(def a-ref (ref 0))

(dosync
  (ref-set a-ref 1))

(defmacro sync-set [r v]
  (list 'dosync
    (list 'ref-set r v)))

(sync-set a-ref 1)


(defn unless [test then]
  (if (not test)
    then))

(defn exhibits-oddity? [x]
  (unless (even? x)
    (println "Very odd, indeed!")))

(exhibits-oddity? 11)
;Very odd, indeed!
;=> nil

(defn unless [test then-thunk]
  (if (not test)
    (then-thunk)))

(defn exhibits-oddity? [x]
  (unless (even? x)
    #(println "Rather odd!")))

(exhibits-oddity? 11)
;Rather odd!
;=> nil
(exhibits-oddity? 10)
;=> nil

(defmacro unless [test then]
  (list 'if (list 'not test)
    then))

(defn exhibits-oddity? [x]
  (unless (even? x)
    (println "Very odd, indeed!")))

(macroexpand
  '(unless (even? x) (println "Very odd, indeed!")))        ; Note we quote the argument
;=> (if (not (even? x)) (println "Very odd, indeed!"))

(defmacro unless [test then]
  `(if (not ~test)
     ~then))

(defmacro unless [test & exprs]
  `(if (not ~test)
     (do ~exprs)))

(macroexpand-1 '(unless (even? x)
                  (println "Odd!")
                  (println "Very odd!")))
;=> (if (clojure.core/not (even? x))
;     (do ((println "Odd!") (println "Very odd!"))))

(defmacro unless [test & exprs]
  `(if (not ~test)
     (do ~@exprs)))


(defmacro def-logged-fn [fn-name args & body]
  `(defn ~fn-name ~args
     (let [now (System/currentTimeMillis)]
       (println "[" now "] Call to" (str (var ~fn-name)))
       ~@body)))

#_(def-logged-fn printname [name]
  (println "hi" name))
;CompilerException java.lang.RuntimeException: Can't let qualified name: clj-in-act.ch7/now, compiling:(NO_SOURCE_PATH:1:1)


(macroexpand-1 '(def-logged-fn printname [name]
                  (println "hi" name)))
;=> (clojure.core/defn printname [name]
;     (clojure.core/let [clj-in-act.ch7/now (java.lang.System/currentTimeMillis)]
;       (clojure.core/println "[" clj-in-act.ch7/now ":] Call to"
;         (clojure.core/str (var printname)))
;       (println "hi" name)))


(defmacro def-logged-fn [fn-name args & body]
  `(defn ~fn-name ~args
     (let [now# (System/currentTimeMillis)]
       (println "[" now# "] Call to" (str (var ~fn-name)))
       ~@body)))

(def-logged-fn printname [name]
  (println "hi" name))

(printname "deepthi")
;[ 1259955655338 ] Call to #'clj-in-act.ch7/printname
;hi deepthi
;=> nil


;;; 7.2 Macros from within Clojure

#_(defmacro comment [& body])

#_(defmacro declare [& names]
  `(do
     ~@(map #(list 'def %) names)))


(macroexpand '(declare add multiply subtract divide))
;=> (do
;     (def add)
;     (def multiply)
;     (def subtract)
;     (def divide))


#_(defmacro defonce [name expr]
  `(let [v# (def ~name)]
     (when-not (.hasRoot v#)
       (def ~name ~expr))))

#_(defmacro and
  ([] true)
  ([x] x)
  ([x & next]
   `(let [and# ~x]
      (if and# (and ~@next) and#))))


(macroexpand '(and (even? x) (> x 50) (< x 500)))
;=> (let* [and__4357__auto__ (even? x)]
;     (if and__4357__auto__
;       (clojure.core/and (> x 50) (< x 500))
;       and__4357__auto__) )


(time (* 1331 13531))
;"Elapsed time: 0.04 msecs"
;=> 18009761

#_(defmacro time [expr]
  `(let [start# (. System (nanoTime))
         ret# ~expr]
     (prn
       (str "Elapsed time: "
         (/ (double (- (. System (nanoTime)) start#)) 1000000.0)
         " msecs"))
     ret#))


;;; 7.3 Writing your own macros


(defmacro infix [expr]
  (let [[left op right] expr]
    (list op left right)))

(defmacro randomly [& exprs]
  (let [len (count exprs)
        index (rand-int len)
        conditions (map #(list '= index %) (range len))]
    `(cond ~@(interleave conditions exprs))))

(randomly (println "amit") (println "deepthi") (println "adi"))
;adi
;=> nil

(randomly (println "amit") (println "deepthi") (println "adi"))
;deepthi
;=> nil

(randomly (println "amit") (println "deepthi") (println "adi"))
;adi
;=> nil

(macroexpand-1
  '(randomly (println "amit") (println "deepthi") (println "adi")))
;=> (clojure.core/cond
;     (= 0 0) (println "amit")
;     (= 0 1) (println "deepthi")
;     (= 0 2) (println "adi"))

(defmacro randomly-2 [& exprs]
  (nth exprs (rand-int (count exprs))))

(defmacro randomly-2 [& exprs]
  (let [c (count exprs)]
    `(case (rand-int ~c) ~@(interleave (range c) exprs))))

(defn check-credentials [username password]
  true)

(defn login-user [request]
  (let [username (:username request)
        password (:password request)]
    (if (check-credentials username password)
      (str "Welcome back, " username ", " password " is correct!")
      (str "Login failed!"))))

(def request {:username "amit" :password "123456"})
;=> #'clj-in-act.ch7/request
(login-user request)
;=> "Welcome back, amit, 123456 is correct!"

(defmacro defwebmethod [name args & exprs]
  `(defn ~name [{:keys ~args}]
     ~@exprs))

(defwebmethod login-user [username password]
  (if (check-credentials username password)
    (str "Welcome, " username ", " password " is still correct!")
    (str "Login failed!")))

(login-user request)
;=> "Welcome, amit, 123456 is still correct!"


(defmacro defnn [fname [& names] & body]
  (let [ks {:keys (vec names)}]
    `(defn ~fname [& {:as arg-map#}]
       (let [~ks arg-map#]
         ~@body))))

(defnn print-details [name salary start-date]
  (println "Name:" name)
  (println "Salary:" salary)
  (println "Started on:" start-date))
;=> #'clj-in-act.ch7/print-details

(macroexpand
  '(defnn print-details [name salary start-date]
     (println "Name:" name)
     (println "Salary:" salary)
     (println "Started on:" start-date)))

;=> (def print-details
;     (clojure.core/fn
;       ([& {:as arg-map__2045__auto__}]
;        (clojure.core/let [{:keys [name salary start-date]}
;                           arg-map__2045__auto__]
;         (println "Name:" name)
;         (println "Salary:" salary)
;         (println "Started on:" start-date)))))


(print-details :start-date "10/22/2009" :name "Rob" :salary 1000000)
;Name: Rob
;Salary: 1000000
;Started on: 10/22/2009


(defmacro assert-true [test-expr]
  (let [[operator lhs rhs] test-expr]
    `(let [rhsv# ~rhs ret# ~test-expr]
       (if-not ret#
         (throw (RuntimeException.
                  (str '~lhs " is not " '~operator " " rhsv#)))
         true))))

(macroexpand-1 '(assert-true (>= (* 2 4) (/ 18 2))))
;=> (clojure.core/let [lhsv__11965__auto__ (* 2 4)
;                      rhsv__11966__auto__ (/ 18 2)
;                      ret__11967__auto__ (>= (* 2 4) (/ 18 2))]
;      (clojure.core/if-not ret__11967__auto__
;        (throw (java.lang.RuntimeException.
;          (clojure.core/str (quote (* 2 4))
;              " is not " (quote >=) " " rhsv__11966__auto__)))
;      true))


(assert-true (= (* 2 4) (/ 16 2)))
;=> true
(assert-true (< (* 2 4) (/ 18 2)))
;=> true

#_(assert-true (>= (* 2 4) (/ 18 2)))
;=> RuntimeException (* 2 4) is not >= 9


(defmacro assert-true [test-expr]
  (if-not (= 3 (count test-expr))
    (throw (RuntimeException.
             "Argument must be of the form
                   (operator test-expr expected-expr)")))
  (if-not (some #{(first test-expr)} '(< > <= >= = not=))
    (throw (RuntimeException.
             "operator must be one of < > <= >= = not=")))
  (let [[operator lhs rhs] test-expr]
    `(let [rhsv# ~rhs ret# ~test-expr]
       (if-not ret#
         (throw (RuntimeException.
                  (str '~lhs " is not " '~operator " " rhsv#)))
         true))))

#_(assert-true (>= (* 2 4) (/ 18 2) (+ 2 5)))
;=> RuntimeException Argument must be of the form
;   (operator test-expr expected-expr)


#_(assert-true (<> (* 2 4) (/ 16 2)))
;=> RuntimeException operator must be one of < > <= >= = not=



