(ns clj-in-act.ch8
  "More on functional programming")


;;; 8.1 Using higher-order functions

(defn square [x]
  (* x x))

(defn square-all [numbers]
  (if (empty? numbers)
    nil                                                     ; Idiomatic equivalent to an empty list. (See discussion in 8.1.1.)
    (cons (square (first numbers))
      (square-all (rest numbers)))))

(square-all [1 2 3 4 5 6])
;=> (1 4 9 16 25 36)


(defn cube [x]
  (* x x x))

(defn cube-all [numbers]
  (if (empty? numbers)
    ()
    (cons (cube (first numbers))
      (cube-all (rest numbers)))))

(cube-all [1 2 3 4 5 6])
;=> (1 8 27 64 125 216)

(defn do-to-all [f numbers]
  (if (empty? numbers)
    ()
    (cons (f (first numbers))
      (do-to-all f (rest numbers)))))

(do-to-all square [1 2 3 4 5 6])
;=> (1 4 9 16 25 36)
(do-to-all cube [1 2 3 4 5 6])
;=> (1 8 27 64 125 216)

#_(do-to-all square (range 11000))
;StackOverflowError   clojure.lang.Numbers$LongOps.multiply (Numbers.java:459)

(defn do-to-all [f numbers]
  (lazy-seq
    (if (empty? numbers)
      ()
      (cons (f (first numbers))
        (do-to-all f (rest numbers))))))

(take 10 (drop 10000 (do-to-all square (range 11000))))
;=> (100000000 100020001 100040004 100060009 100080016 100100025 100120036
;    100140049 100160064 100180081)

(take 10 (drop 10000 (do-to-all square (range 11000))))
;=> (100000000 100020001 100040004 100060009 100080016 100100025 100120036
;    100140049 100160064 100180081)

(defn total-of [numbers]
  (loop [nums numbers sum 0]
    (if (empty? nums)
      sum
      (recur (rest nums) (+ sum (first nums))))))

(total-of [5 7 9 3 4 1 2 8])
;=> 39

(defn larger-of [x y]
  (if (> x y) x y))

(defn largest-of [numbers]
  (loop [l numbers candidate (first numbers)]
    (if (empty? l)
      candidate
      (recur (rest l) (larger-of candidate (first l))))))

(largest-of [5 7 9 3 4 1 2 8])
;=> 9
(largest-of [])
;=> nil

(defn compute-across [func elements value]
  (if (empty? elements)
    value
    (recur func (rest elements) (func value (first elements)))))

(defn total-of [numbers]
  (compute-across + numbers 0))

(defn largest-of [numbers]
  (compute-across larger-of numbers (first numbers)))

(total-of [5 7 9 3 4 1 2 8])
;=> 39
(largest-of [5 7 9 3 4 1 2 8])
;=> 9

(defn all-greater-than [threshold numbers]
  (compute-across #(if (> %2 threshold) (conj %1 %2) %1) numbers []))

(all-greater-than 5 [5 7 9 3 4 1 2 8])
;=> [7 9 8]

(defn all-greater-than [threshold numbers]
  (reduce #(if (> %2 threshold) (conj %1 %2) %1) [] numbers))

(all-greater-than 5 [5 7 9 3 4 1 2 8])
;=> [7 9 8]

(defn all-lesser-than [threshold numbers]
  (compute-across #(if (< %2 threshold) (conj %1 %2) %1) numbers []))

(all-lesser-than 5 [5 7 9 3 4 1 2 8])
;=> [3 4 1 2]

(defn select-if [pred elements]
  (compute-across #(if (pred %2) (conj %1 %2) %1) elements []))

(select-if odd? [5 7 9 3 4 1 2 8])
;=> [5 7 9 3 1]

(defn all-lesser-than [threshold numbers]
  (select-if #(< % threshold) numbers))

(filter odd? [5 7 9 3 4 1 2 8])
;=> (5 7 9 3 1)

;;; 8.2 Partial application

(defn price-with-tax [tax-rate amount]
  (->> (/ tax-rate 100)
    (+ 1)
    (* amount)))

(price-with-tax 9.5M 100)
;=> 109.500M

(defn with-california-taxes [prices]
  (map #(price-with-tax 9.25M %) prices))

(def prices [100 200 300 400 500])
;=> #'clj-in-act.ch8/prices
(with-california-taxes prices)
;=> (109.2500M 218.5000M 327.7500M 437.0000M 546.2500M)

(defn price-with-ca-tax [price]
  (price-with-tax 9.25M price))

(defn price-with-ny-tax [price]
  (price-with-tax 8.0M price))

(defn price-calculator-for-tax [state-tax]
  (fn [price]
    (price-with-tax state-tax price)))

(def price-with-ca-tax (price-calculator-for-tax 9.25M))
(def price-with-ny-tax (price-calculator-for-tax 8.0M))

(defn of-n-args [a b c d e]
  (str a b c d e))

(defn of-k-args [d e]
  (of-n-args 1 2 3 d e))

(of-k-args \a \b)
;=> "123ab"

(defn partially-applied [of-n-args & n-minus-k-args]
  (fn [& k-args]
    (apply of-n-args (concat n-minus-k-args k-args))))

(def of-2-args (partially-applied of-n-args \a \b \c))
;=> #'clj-in-act.ch8/of-2-args
(def of-3-args (partially-applied of-n-args \a \b))
;=> #'clj-in-act.ch8/of-3-args

(of-2-args 4 5)
;=> "abc45"
(of-3-args 3 4 5)
;=> "ab345"

(def of-2-args (partial of-n-args \a \b \c))
;=> #'clj-in-act.ch8/of-2-args
(def of-3-args (partial of-n-args \a \b))
;=> #'clj-in-act.ch8/of-3-args

(of-2-args 4 5)
;=> "abc45"
(of-3-args 3 4 5)
;=> "ab345"

(defn select-into-if [container pred elements]
  (compute-across #(if (pred %2) (conj %1 %2) %1) elements container))

(def numbers [4 9 5 7 6 3 8])

(select-into-if [] #(< % 7) numbers)
;=> [4 5 6 3]

(select-into-if () #(< % 7) numbers)
;=> (3 6 5 4)

(def select-up (partial select-into-if []))

(def select-down (partial select-into-if ()))

(select-up #(< % 9) [5 3 9 6 8])
;=> [5 3 6 8]
(select-down #(< % 9) [5 3 9 6 8])
;=> (8 6 3 5)

;;; 8.3 Closures


(defn adder [num1 num2]
  (let [x (+ num1 num2)]
    (fn [y]
      ;; This `x` is a free variable: its value will come from outside scope of
      ;; the function that uses it.
      (+ x y))))

(def add-5 (adder 2 3))
;=> #'clj-in-act.ch8/add-5
(add-5 10)
;=> 15

#_(let [x 1
      y 0]
  (/ x y))
;ArithmeticException Divide by zero  clojure.lang.Numbers.divide (Numbers.java:156)


(let [x 1
      y 0]
  (try
    (/ x y)
    (catch Exception e (println (.getMessage e)))))
;Divide by zero
;=> nil

(defn try-catch [the-try the-catch]
  (try
    (the-try)
    (catch Exception e (the-catch e))))


(let [x 1
      y 0]
  (try-catch #(/ x y)
    #(println (.getMessage %))))

(defn new-user [login password email]
  (fn [a]
    (case a
      :login login
      :password password
      :email email
      nil)))

(def arjun (new-user "arjun" "secret" "arjun@zololabs.com"))
;=> #'clj-in-act.ch8/arjun
(arjun :login)
;=> "arjun"
(arjun :password)
;=> "secret"
(arjun :email)
;=> "arjun@zololabs.com"
(arjun :name)
;=> nil


(defn new-user [login password email]
  (fn [a]
    (case a
      :login login
      :email email
      :password-hash (hash password)                        ; The inner function can access password,
      nil)))
;=> #'clj-in-act.ch8/new-user
(def arjun (new-user "arjun" "secret" "arjun@zololabs.com"))
;=> #'clj-in-act.ch8/arjun
(arjun :password)                                           ; but does not ever return the password to callers.
;=> nil
(arjun :password-hash)                                      ; This enables information hiding: we can use password
; without letting anyone else see it.
;=> 1614358358


(defn new-user [login password email]
  (fn [a & args]
    (case a
      :login login
      :email email
      :authenticate (= password (first args)))))

(def adi (new-user "adi" "secret" "adi@currylogic.com"))
;=> #'clj-in-act.ch8/adi
(adi :authenticate "blah")
;=> false
(adi :authenticate "secret")
;=> true


(defn new-class [class-name]
  (fn [command & args]
    (case command
      :name (name class-name))))

(defmacro defclass [class-name]
  `(def ~class-name (new-class '~class-name)))


(defclass Person)
;=> #'clj-in-act.ch8/Person
(Person :name)
;=> "Person"

(def some-class Person)
;=> #'clj-in-act.ch8/some-class
(some-class :name)
;=> "Person"

(defn new-object [klass]
  (fn [command & args]
    (case command
      :class klass)))

(def cindy (new-object Person))

(new-object Person)
;=> #<clj-in-act.ch8$new_object$fn__2259 clj-in-act.ch8$new_object$fn__2259@1f106fec>

((cindy :class) :name)
;=> "Person"

(defn new-object [klass]
  (fn [command & args]
    (case command
      :class klass
      :class-name (klass :name))))

(def cindy (new-object Person))
;=> #'clj-in-act.ch8/cindy
(cindy :class-name)
;=> "Person"

(defn new-class [class-name]
  (fn klass [command & args]
    (case command
      :name (name class-name)
      :new (new-object klass))))

(defclass Person)
;=> #'clj-in-act.ch8/Person
(def nancy (Person :new))
;=> #'clj-in-act.ch8/nancy
(nancy :class-name)
;=> "Person"

(defn new-object [klass]
  (let [state (ref {})]
    (fn [command & args]
      (case command
        :class klass
        :class-name (klass :name)
        :set! (let [[k v] args]
                (dosync (alter state assoc k v))
                nil)
        :get (let [[key] args]
               (@state key))))))

(def nancy (Person :new))
;=> #'clj-in-act.ch8/nancy
(nancy :set! :name "Nancy Warhol")
;=> nil
(nancy :get :name)
;=> "Nancy Warhol"

(nancy :set! :name "Nancy Drew")
;=> nil
(nancy :get :name)
;=> "Nancy Drew"

(defn method-spec [sexpr]
  (let [name (keyword (second sexpr))
        body (next sexpr)]
    [name (conj body 'fn)]))

(method-spec '(method age [] (* 2 10)))
;=> [:age (fn age [] (* 2 10))]

(defn method-specs [sexprs]
  (->> sexprs
    (filter #(= 'method (first %)))
    (mapcat method-spec)
    (apply hash-map)))

(method-specs '((method age [] (* 2 10))
                 (method greet [visitor] (str "Hello there, " visitor))))
;=> {:age (fn age [] (* 2 10)),
;    :greet (fn greet [visitor] (str "Hello there, " visitor))}

(defn new-class [class-name methods]
  (fn klass [command & args]
    (case command
      :name (name class-name)
      :new (new-object klass))))

(defmacro defclass [class-name & specs]
  (let [fns (or (method-specs specs) {})]
    `(def ~class-name (new-class '~class-name ~fns))))

(defclass Person
  (method age [] (* 2 10))
  (method greet [visitor] (str "Hello there, " visitor)))
;=> #'clj-in-act.ch8/Person


(defn find-method [method-name instance-methods]
  (instance-methods method-name))

(defn new-class [class-name methods]
  (fn klass [command & args]
    (case command
      :name (name class-name)
      :new (new-object klass)
      :method (let [[method-name] args]
                (find-method method-name methods)))))

;;; We must redef Person to pick up the changed macros.
(defclass Person
  (method age [] (* 2 10))
  (method greet [visitor] (str "Hello there, " visitor)))
;=> #'clj-in-act.ch8/Person


(Person :method :age)
;=> #<clj-in-act.ch8$age clj-in-act.ch8$age@42443032>

((Person :method :age))
;=> 20


(defn new-object [klass]
  (let [state (ref {})]
    (fn [command & args]
      (case command
        :class klass
        :class-name (klass :name)
        :set! (let [[k v] args]
                (dosync (alter state assoc k v))
                nil)
        :get (let [[key] args]
               (@state key))
        (if-let [method (klass :method command)]
          (apply method args)
          (throw (RuntimeException.
                   (str "Unable to respond to " command))))))))

;;; We must redef Person to pick up the changed macros.
(defclass Person
  (method age [] (* 2 10))
  (method greet [visitor] (str "Hello there, " visitor)))
;=> #'clj-in-act.ch8/Person


(def shelly (Person :new))
;=> #'clj-in-act.ch8/shelly
(shelly :age)
;=> 20
(shelly :greet "Nancy")
;=> "Hello there, Nancy"


(declare ^:dynamic this)

(defn new-object [klass]
  (let [state (ref {})]
    (fn thiz [command & args]
      (case command
        :class klass
        :class-name (klass :name)
        :set! (let [[k v] args]
                (dosync (alter state assoc k v))
                nil)
        :get (let [[key] args] (@state key))
        (let [method (klass :method command)]
          (if-not method
            (throw (RuntimeException.
                     (str "Unable to respond to " command))))
          (binding [this thiz]
            (apply method args)))))))


(defclass Person
  (method age [] (* 2 10))
  (method about [diff]
    (str "I was born about " (+ diff (this :age)) " years ago")))

(def shelly (Person :new))
;=> #'clj-in-act.ch8/shelly
(shelly :about 2)
;=> "I was born about 22 years ago"


(defn parent-class-spec [sexprs]
  (let [extends-spec (filter #(= 'extends (first %)) sexprs)
        extends (first extends-spec)]
    (if (empty? extends)
      'OBJECT
      (last extends))))

(parent-class-spec '((extends Person)
                      (method age [] (* 2 9))))
;=> Person

(var map)
;=> #'clojure.core/map

(defmacro defclass [class-name & specs]
  (let [parent-class (parent-class-spec specs)
        fns (or (method-specs specs) {})]
    `(def ~class-name (new-class '~class-name #'~parent-class ~fns))))

(defn new-class [class-name parent methods]
  (fn klass [command & args]
    (case command
      :name (name class-name)
      :parent parent
      :new (new-object klass)
      :method (let [[method-name] args]
                (find-method method-name methods)))))

(def OBJECT (new-class :OBJECT nil {}))

(defclass Person
  (method age [] (* 2 10))
  (method about [diff]
    (str "I was born about " (+ diff (this :age)) " years ago")))
;=> #'clj-in-act.ch8/Person

(defclass Woman
  (extends Person)
  (method greet [v] (str "Hello, " v))
  (method age [] (* 2 9)))
;=> #'clj-in-act.ch8/Woman

(def donna (Woman :new))
;=> #'clj-in-act.ch8/donna
(donna :greet "Shelly")
;=> "Hello, Shelly"
(donna :age)
;=> 18
#_(donna :about 3)
;RuntimeException Unable to respond to :about  clj-in-act.ch8/new-object/thiz--2733 (NO_SOURCE_PATH:1:1)

(defn find-method [method-name klass]
  (or ((klass :methods) method-name)
    (if-not (= #'OBJECT klass)
      (find-method method-name (klass :parent)))))

(defn new-class [class-name parent methods]
  (fn klass [command & args]
    (case command
      :name (name class-name)
      :parent parent
      :new (new-object klass)
      :methods methods
      :method (let [[method-name] args]
                (find-method method-name klass)))))

(defclass Person
  (method age [] (* 2 10))
  (method about [diff]
    (str "I was born about " (+ diff (this :age)) " years ago")))
;=> #'clj-in-act.ch8/Person

(defclass Woman
  (extends Person)
  (method greet [v] (str "Hello, " v))
  (method age [] (* 2 9)))
;=> #'clj-in-act.ch8/Woman

(def donna (Woman :new))

(donna :about 3)
;=> "I was born about 21 years ago"
